"""
pdfplucker - a powerful wrapper for the Docling library
"""

__version__ = "0.2.5"

from src.processor import process_batch, process_pdf, create_converter
from src.utils import format_result, link_subtitles
__all__ = [
    "process_batch",
    "process_pdf",
    "create_converter",
    "format_result",
    "link_subtitles"
]


from docling.datamodel.document import ConversionResult
from docling_core.types.doc import (
    PictureItem,
    TableItem,
    TextItem,
    DocItemLabel,
)
from typing import TypedDict, List, Dict, Any

class Data(TypedDict):
    metadata: Dict[str, Any]
    sections: List[Dict[str, Any]]
    images: List[Dict[str, Any]]
    tables: List[Dict[str, Any]]
    subtitles: List[Dict[str, Any]]

def format_result(conv: ConversionResult, data: Data, filename: str, image_path: str) -> None:
    ''' Uses the docling document to format a readable JSON result '''

    collecting = None
    counter = 0

    for item, _, in conv.document.iterate_items():
        if isinstance(item, TextItem):
            if item.label == DocItemLabel.SECTION_HEADER:
                if collecting is not None:
                    data['sections'].append(collecting)
                collecting = {'title': item.text, 'text': ''}
            else:
                if collecting is not None:
                    collecting['text'] += '\n' + item.text if collecting['text'] else item.text
        elif isinstance(item, TableItem):
            table = item.export_to_dataframe()
            data['tables'].append({
                'self_ref' : item.self_ref,
                'subtitle' : '',
                'table' : table.to_dict()
            })
        elif isinstance(item, PictureItem):
            classification = None
            if item.annotations:
                for annotation in item.annotations:
                    if annotation.kind == 'classification':
                        # Find the classification with the highest confidence
                        best_class = max(
                            annotation.predicted_classes,
                            key=lambda cls: cls.confidence
                        )
                        classification = {
                            'class_name': best_class.class_name,
                            'confidence': best_class.confidence
                        }
                        break

            image_filename = (image_path  / f"{filename}_{counter}.png")
            with image_filename.open('wb') as f:
                item.get_image(conv.document).save(f, "PNG")
            data['images'].append({
                'ref': f"{filename}_{counter}.png",
                'self_ref' : item.self_ref,
                'classification' : classification,
                'subtitle' : ''
            })
            counter += 1

    # Collecting the results after all iterations
    if collecting is not None:
        data['sections'].append(collecting)
    for text in conv.document.texts:
        if text.label == 'caption':
            data['subtitles'].append({
                'ref' : text.parent.cref,
                'text' : text.text
            })

def link_subtitles(data: Data) -> None:
    ''' 
    Try associating subtitles with images or tables
    based on refs saved on the Docling document
    '''
    images = data.get('images', [])
    tables = data.get('tables', [])
    subtitles = data.get('subtitles', [])

    # iterate over all images and all subtitles
    for img in images:
        self_ref = img.get('self_ref')
        if self_ref is None:
            continue
        
        # For each image, try to find a ref equals to the image self ref
        for sub in subtitles[:]:
            ref = sub.get('ref')
            if self_ref == ref:
                img['subtitle'] = sub.get('text', '')
                img.pop('item', None)
                # remove the subtitle to be sure
                subtitles.remove(sub)
                break 

    # iterate over all tables and the remaining subtitles
    for tab in tables:
        self_ref = tab.get('self_ref')
        if self_ref is None:
            continue

        # For each table, try to find a ref equals to the table self ref
        for sub in subtitles[:]:
            ref = sub.get('ref')
            if self_ref == ref:
                tab['subtitle'] = sub.get('text', '')
                subtitles.remove(sub)
                break

    for sub in subtitles:
        data['images'].append({
            'ref': sub.get('ref'),
            'self_ref' : None,
            'subtitle' : sub.get('text', '')
        })

    data.pop('subtitles')
    
    del subtitles, tables, images


import os
import gc
import json
import fitz
import time
import multiprocessing
from pathlib import Path
from src.utils import format_result, link_subtitles, Data
from concurrent.futures import ProcessPoolExecutor, as_completed, TimeoutError
from docling.document_converter import DocumentConverter, PdfFormatOption
from docling.datamodel.base_models import InputFormat
from docling.datamodel.document import ConversionResult
from docling_core.types.doc import ImageRefMode
from docling.datamodel.pipeline_options import (
    AcceleratorDevice,
    AcceleratorOptions,
    PdfPipelineOptions,
    RapidOcrOptions
)

def create_converter(device : str = 'CPU', num_threads : int = 4, ocr_lang: list = ['es', 'pt'], force_ocr: bool = False) -> DocumentConverter:
    ''' Create a DocumentConverter object with the pipeline options configured''' 

    pipeline_options = PdfPipelineOptions()
    pipeline_options.do_ocr = True
    pipeline_options.do_table_structure = True
    pipeline_options.table_structure_options.do_cell_matching = True 
    pipeline_options.ocr_options.lang = ocr_lang
    pipeline_options.generate_picture_images = True
    pipeline_options.do_picture_classification = True
    pipeline_options.images_scale = 2.0
    if force_ocr:
    # Rapid OCR or Easy OCr
        ocr_options = RapidOcrOptions(force_full_page_ocr=True)
        pipeline_options.ocr_options = ocr_options
    # Device acceleration
    device_type = AcceleratorDevice.CUDA if device.upper() == 'CUDA' else AcceleratorDevice.CPU if device.upper() == 'CPU' else AcceleratorDevice.AUTO if device.upper() == 'AUTO' else AcceleratorDevice.AUTO
    pipeline_options.accelerator_options = AcceleratorOptions(num_threads=num_threads, device=device_type)

    return DocumentConverter(
        format_options={
            InputFormat.PDF: PdfFormatOption(pipeline_options=pipeline_options)
        }
    )

def process_pdf(source: str, output: str, image_path: str | None, doc_converter: DocumentConverter, separate_folders: bool | None = False, markdown: bool = False) -> bool:
    """Function to process a single PDF file utilizing Docling"""

    filename = os.path.basename(source)
    base_filename = os.path.splitext(filename)[0]

    try:
        if separate_folders:
            specific_folder = os.path.join(output, base_filename)
            result = os.path.join(specific_folder, f"{base_filename}.json")
            image_folder = os.path.join(specific_folder, "images")
            os.makedirs(specific_folder, exist_ok=True)
            os.makedirs(image_folder, exist_ok=True)
        else:
            result = os.path.join(output, f"{base_filename}.json")
            image_folder = image_path

        image_folder_path = Path(image_folder)

        print(f"\033[33mProcessing: {filename}\033[0m")

        data: Data = {
            "metadata": {},
            "sections" : [],
            "images": [],
            "tables": [],
            "subtitles" : []
        }

        with fitz.open(source) as doc:
            data["metadata"] = doc.metadata
            data["metadata"]["filename"] = filename

        # Converting the source file to a Docling document
        conv: ConversionResult = doc_converter.convert(source)
        format_result(conv, data, base_filename, image_folder_path)
        link_subtitles(data)

        # Save Markdown if asked
        if markdown:
            md_filename = result.replace('.json', '.md')
            conv.document.save_as_markdown(md_filename, image_mode=ImageRefMode.EMBEDDED)
        
        # Save JSON
        with open(result, 'w', encoding='utf-8') as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

        return True
    
    except (fitz.FileDataError, fitz.EmptyFileError) as e:
        print(f"\033[31mFailed to process {filename}: {e}\033[0m")
        return False
    except IOError as e:
        print(f"\033[31mI/O error while processing: {e}\033[0m")
        return False
    except Exception as e:
        print(f"\033[31mAn error has occurred: {e}\033[0m") 
        return False
    
    finally:
        try:
            if 'conv' in locals():
                del conv
            if 'data' in locals():
                del data
            gc.collect()
        except Exception as e:
            print(f"\033[31mAn error has occurred while trying to free memory: {e}\033[0m")

def _worker(source, output, image_path, doc_converter, separate_folders, markdown, queue):
    try:
        success = process_pdf(source, output, image_path, doc_converter, separate_folders, markdown)
        queue.put(success)
    except Exception as e:
        print(f"\033[31mNon treated error at _worker: {e}\033[0m")
        queue.put(False)

def process_with_timeout(source: str, output: str, image_path: str, doc_converter: DocumentConverter, separate_folders: bool = False, timeout: int = 600, markdown:bool = False):
    """ Process a single PDF with safety timeout """

    if multiprocessing.get_start_method() != 'spawn':
        multiprocessing.set_start_method('spawn', force=True)

    queue = multiprocessing.Queue()

    process = multiprocessing.Process(
        target=_worker,
        args=(source, output, image_path, doc_converter, separate_folders, markdown, queue) 
        )
    
    process.start()
    process.join(timeout)

    if process.is_alive():
        filename = os.path.basename(source)
        print(f"\033[31mTimeout after {timeout}s! Killing process for {filename}\033[0m")
        process.terminate()
        process.join()
        return False
    
    try:
        if not queue.empty():
            return queue.get()
        return False
    except Exception:
        return False

def process_batch(source: str | Path, output: str, image_path: str | None, separate_folders: bool = False, max_workers: int = 4, timeout:int = 600, device: str = 'AUTO', markdown: bool = False, force_ocr: bool = False) -> dict:
    """ Process a batch of PDFs at a time in paralel """

    source = Path(source)
    output = Path(output)

    # path config
    if not separate_folders and image_path is None:
        image_path = output / "images"
    elif image_path is not None:
        image_path = Path(image_path)

    # create the doc_converter
    doc_converter = create_converter(device=device, num_threads=max_workers, force_ocr=force_ocr)

    # create the list of files
    pdf_files = []
    if source.is_dir():
        for file in os.listdir(source):
            if file.lower().endswith('.pdf'):
                pdf_files.append(os.path.join(source, file))
    else:
        # Redundancy for safety
        if source.suffix.lower() == '.pdf':
            pdf_files.append(str(source))

    total_docs = len(pdf_files)
    print(f"\033[33mTotal amount of documents to be processed: {total_docs}\033[0m")

    # create the metrics
    metrics = {
        'initial_time' : time.time(),
        'elapsed_time': 0,
        'total_docs' : total_docs,
        'processed_docs': 0,
        'failed_docs': 0,
        'timeout_docs': 0,
        'success_rate' : 0.1,
        'fails': []
    }

    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        futures  = {
            executor.submit(
                process_with_timeout,
                pdf_file,
                str(output),
                str(image_path),
                doc_converter,
                separate_folders,
                timeout,
                markdown
            ) : pdf_file for pdf_file in pdf_files
        }

        for future in as_completed(futures):
            pdf_file = futures[future]
            try: 
                success = future.result()
                if success:
                    metrics['processed_docs'] += 1
                    print(f"\033[32mSuccessfuly processed {os.path.basename(pdf_file)}\033[0m")
                else:
                    metrics['processed_docs'] += 1
                    metrics['failed_docs'] += 1
                    metrics['fails'].append({
                        'file' : pdf_file,
                        'error': 'unknown'
                    })
                    print(f"\033[31m{pdf_file} failed\033[0m")
            except TimeoutError:
                print(f"\033[31m{os.path.basename(pdf_file)} reached time limit! Killing process\033[0m")
                metrics['timeout_docs'] += 1
                metrics['fails'].append({
                    'file' : pdf_file,
                    'error': "Timeout reached"
                })

            except Exception as e:
                print(f"\033[31mAn error has occurred while processing {os.path.basename(pdf_file)}: {e}\033[0m")
                metrics['failed_docs'] += 1
                metrics['fails'].append({
                    'file' : pdf_file,
                    'error': str(e)
                })

    # Conclude metrics
    total_time = time.time() - metrics['initial_time']
    metrics['elapsed_time'] = total_time
    metrics['success_rate'] = (metrics['failed_docs'] + metrics['timeout_docs']) / total_docs * 100

    print(f"\033[32mParsing process has finished!\033[0m")
    return metrics


import sys
import argparse
import time
import json

from pathlib import Path
from src.processor import process_batch, process_pdf, create_converter

def create_parser():
    '''
    Create the argument parser for the command line interface.
    '''

    parser = argparse.ArgumentParser(
        description='Docling wrapper for extracting PDF information.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # Arguments
    parser.add_argument(
        '-s', '--source',
        help='Path to the PDF files (directory or just a single file)'
    )

    parser.add_argument(
        '-o', '--output',
        help='Path to save the processed information',
        default='./results'
    )

    parser.add_argument(
        '-f', '--folder-separation',
        action='store_true',
        help='Create separate folders for each PDF'
    )

    parser.add_argument(
        '-i', '--images',
        help='Path to save the extracted images (ignore if --folder-separation is active)'
    )

    parser.add_argument(
        '-t', '--timeout',
        type=int,
        default=600, # 10 minutes
        help='Time limit in seconds for processing each PDF'
    )

    parser.add_argument(
        '-w', '--workers',
        type=int,
        default=4,
        help='Number of paralels processes (threads)'
    )

    # For forcing ocr use
    parser.add_argument(
        '-ocr', '--force-ocr',
        action='store_true',
        help='Force OCR even if the PDF is not scanned'
    )

    parser.add_argument(
        '-d', '--device',
        choices=['CPU', 'CUDA', 'AUTO'],
        default='AUTO',
        help='Type of device for processing (CUDA, CPU or AUTO)'
    )

    parser.add_argument(
        '-m', '--markdown',
        action='store_true',
        help='Export the document in an aditional markdown file'
    )

    return parser

def validate_args(args: argparse.Namespace):
    '''This function check the many arguments needs'''

    # Checking for the source path
    source_path = Path(args.source)
    if not source_path.exists():
        return False, f"Source path not found: {args.source}"
    
    # Checking if source, as a directory, contains PDFs
    if source_path.is_dir():
        pdf_files = list(source_path.glob("*.pdf"))
        if not pdf_files:
            return False, f"No PDF files found: {args.source}"
        
    # Checkging if source, as a file, is a PDF
    elif not source_path.name.lower().endswith('.pdf'):
        return False, f"Source path is neither a PDF, nor a directory: {args.source}"
    
    # Check number of workers
    if args.workers < 1:
        return False, f"Number of workers must be greater than 0: {args.workers}"
    
    # Check timeout
    if args.timeout < 1:
        return False, f"Timeout must be greater than 0: {args.timeout}"
    
    # Check if output path is a directory, and create if necessary
    output_path = Path(args.output)
    if output_path.exists() and not output_path.is_dir():
        return False, f"Output path is not a directory: {args.output}"
    if not output_path.exists():
        output_path.mkdir(parents=True, exist_ok=True)
        print(f"\033[33mWarning: Output path doesn't exist, creating at source directory: {args.output}\033[0m")

    # Check if images path is a directory, and create if necessary
    if args.images:
        images_path = Path(args.images)
        if images_path.exists() and not images_path.is_dir():
            return False, f"Images path is not a directory: {args.images}"
        if not images_path.exists():
            images_path.mkdir(parents=True, exist_ok=True)
            print(f"\033[32mImages path created: {args.images}\033[0m")
    
    # Check if the user wants to use folder separation and images path at the same time
    if args.folder_separation and args.images:
        return False, "Folder separation and images path cannot be used at the same time."
    
    # Check, if not using folder separation, if image_path is set, and create if necessary
    if not args.folder_separation and not args.images:
        print("\033[33mWarning: Images path not set. Using default path.\033[0m")
        images_path = output_path / 'images'
        if not images_path.exists():
            images_path.mkdir(parents=True, exist_ok=True)

    if not args.folder_separation and args.images:
        if not images_path.exists():
            images_path.mkdir(parents=True, exist_ok=True)
            print(f"\033[33mImages path created: {args.images}\033[0m")
    
    return True, None

def process_single_file(args: argparse.Namespace):
    '''Process a single PDF file and save the results'''
    source_path = Path(args.source)
    output_path = Path(args.output)

    if args.folder_separation:
        images_path = output_path / source_path.stem / 'images'
    elif args.images:
        images_path = Path(args.images)
    else:
        images_path = output_path / 'images'
    
    # Create the images path if it doesn't exist
    if not images_path.exists():
        images_path.mkdir(parents=True, exist_ok=True)
        print(f"\033[32mImages path created: {images_path}\033[0m")

    doc_converter = create_converter(
        device=args.device.upper(),
        num_threads=args.workers,
        force_ocr=args.force_ocr,
    )

    start_time = time.time()
    success = process_pdf(
        str(source_path),
        str(output_path),
        str(images_path),
        doc_converter,
        args.folder_separation,
        args.markdown,
    )
    elapsed_time = time.time() - start_time
    if success:
        print(f"\033[32mProcessing completed successfully in {elapsed_time:.2f} seconds\033[0m")
        print("=" * 50)
        print(f"Output path: {output_path}")
        print(f"Images path: {images_path}")
    else:
        print(f"\033[31mProcessing failed\033[0m")
    return success

def main():
    '''Main CLI function'''
    parser = create_parser()

    if len(sys.argv) == 1:
        print("\033[34mPDFPlucker CLI - Docling Wrapper\033[0m")
        print("A tool for extracting information from PDF files.")
        print("Use the `--help` flag to see available options.")
        sys.exit(0)

    args = parser.parse_args()
    # Format the arguments
    for arg_name, arg_value in vars(args).items():
        if isinstance(arg_value, str):
            vars(args)[arg_name] = arg_value.replace('\\', '/') # For Windows compatibility
    # Validate the arguments
    valid_args, error = validate_args(args)
    if not valid_args:
        print(f"\033[91mError: {error}\033[0m")
        sys.exit(1)

    # Print the main information
    print("=" * 50)
    print("\033[34mPDFPlucker CLI - Docling Wrapper\033[0m")
    print("=" * 50)
    print(f"Source path: {args.source}")
    print(f"Output path: {args.output}")
    print(f"Device type: {args.device}")
    print(f"Number of workers: {args.workers}")
    print(f"Force OCR: {'yes' if args.force_ocr else 'no'}")
    print(f"Timeout: {args.timeout} seconds")
    print(f"Save markdown: {'yes' if args.markdown else 'no'}")
    print(f"Folder separation: {'yes' if args.folder_separation else 'no'}")
    print(f"Images path: {args.images if args.images else 'not used'}")
    print("=" * 50)
    print("Starting...")

    # Start the processing
    try:
        if Path(args.source).is_file():
            # Process a single PDF file
            sucess =  process_single_file(args)
            sys.exit(0 if sucess else 1)
        else:
            metrics = process_batch(
                source=args.source,
                output=args.output,
                image_path=args.images,
                separate_folders=args.folder_separation,
                max_workers=args.workers,
                timeout=args.timeout,
                device=args.device.upper(),
                markdown=args.markdown,
                force_ocr=args.force_ocr
            )

        # Save metrics to JSON file
        metrics_path = Path(args.output) / f"{Path(args.source).name}_metrics.json"
        with open(metrics_path, 'w', encoding='utf-8') as f:
            json.dump(metrics, f, indent=4, ensure_ascii=False)

        # Print the metrics
        print("=" * 50)
        print("\033[32mProcessing completed successfully\033[0m")
        print(f"\033[32mMetrics saved to: {metrics_path}\033[0m")      
        print("=" * 50)
        print(f"Total amount of files: {metrics['total_docs']}")
        print(f"Successfully processed: {metrics['processed_docs']}")
        print(f"Failed processes: {metrics['failed_docs'] + metrics['timeout_docs']}")
        print(f"Success rate: {metrics['success_rate']}")
        print(f"Total time elapsed: {metrics['elapsed_time']:.2f} seconds")
        print("=" * 50)
    
    except KeyboardInterrupt:
        print("\033[31mProcess interrupted by user\033[0m")
        sys.exit(1)
    except Exception as e:
        print(f"\033[31mAn error occurred: {e}\033[0m")
        sys.exit(1)

if __name__ == "__main__":
    main()